#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Command-line parsing conveniences """
import argparse
import itertools
import sys
from enum import Enum
from pathlib import Path
from typing import List, Type

# pylint: disable=E0401, E0402, R0913
from pycapo import CapoConfig

from .exceptions import FetchError
from .fetcher_factory import ConfiguredFetcherFactory
from .fetchers import ForceMode
from .interfaces import (
    FetcherFactory,
    FetchPlan,
    FetchProgressReporter,
    FileFetcher,
    LocationReportFileEnumerator,
    SizeCalculator,
)
from .locations import FileLocator, LocatedFile, LocationReport, Locator, ServiceLocator
from .progress import ProgressBarReporter, ProgressNonReporter, VerboseProgressReporter


class RetrievalMode(Enum):
    """
    How we're retrieving a file: via streaming or via direct copy (plugin)
    """

    STREAM = "stream"
    COPY = "copy"


class CLIParam(Enum):
    """Codifies productfetcher's various command-line parameters"""

    SPL = "--product-locator"
    FILE = "--location-file"
    DRY = "--dry-run"
    FORCE = "--force"
    DIRECT_COPY = "--direct-copy"
    STREAMING = "--streaming"
    CONCURRENCY = "--concurrency"
    ONLY_SCIENCE = "--single-product"
    ALSO_ANCILLARIES = "--also-ancillaries"
    ONLY_ANCILLARIES = "--ancillaries-for"
    VERBOSE = "--verbose"
    QUIET = "--quiet"
    PROGRESS = "--progress"


class ScienceAndAncillaryMode(LocationReportFileEnumerator, SizeCalculator):
    """
    Determine whether science products, ancillaries or both should be fetched.
    """


class FetchScienceOnly(ScienceAndAncillaryMode):
    """
    Returns all the normal science files under this report.
    """

    def files(self, location_report: LocationReport) -> List[LocatedFile]:
        return location_report.files

    def aggregate_size(self, report: LocationReport) -> int:
        return report.aggregate_size


class FetchAncillaryOnly(ScienceAndAncillaryMode):
    """
    Returns all the files under all the ancillaries under in this report.
    """

    def __init__(self):
        self.science = FetchScienceOnly()

    def files(self, location_report: LocationReport) -> List[LocatedFile]:
        return list(f for ancillary in location_report.ancillaries.values() for f in self.science.files(ancillary))

    def aggregate_size(self, report: LocationReport) -> int:
        return sum(ancillary.aggregate_size for ancillary in report.ancillaries.values())


class FetchBothScienceAndAncillary(ScienceAndAncillaryMode):
    """
    Returns both the science and ancillary files in this report (i.e. everything)
    """

    def __init__(self):
        self.science = FetchScienceOnly()
        self.ancillary = FetchAncillaryOnly()

    def files(self, location_report: LocationReport) -> List[LocatedFile]:
        return self.science.files(location_report) + self.ancillary.files(location_report)

    def aggregate_size(self, report: LocationReport) -> int:
        return self.science.aggregate_size(report) + self.ancillary.aggregate_size(report)


# Prologue and epilogue for the command line parser.
_PROLOGUE = """Retrieve a product (a science product or an ancillary product)
from the NRAO archive, either by specifying the product's locator or by
providing the path to a product locator report."""


class FetchContext:
    """Handles the various command-line options"""

    def __init__(
        self,
        locators: List[Locator],
        streaming: bool,
        direct_copy: bool,
        dry_run=False,
        force: ForceMode = None,
        science_ancillary: ScienceAndAncillaryMode = None,
        concurrency=1,
        progress_reporter: Type[FetchProgressReporter] = None,
    ):
        self.locators = locators
        self.streaming = streaming
        self.direct_copy = direct_copy
        self.dry_run = dry_run
        self.force = force
        self.concurrency = concurrency
        self.science_ancillary = science_ancillary
        self.progress_reporter = progress_reporter
        self.aggregate_size = 0

    def generate_plan(self) -> FetchPlan:
        """
        Generate the fetch plan for the entire product.

        :return: a plan
        """
        # Get the location report
        reports = [locator.locate() for locator in self.locators]

        # Using the arguments we've parsed, generate the fetcher factory
        factory = ConfiguredFetcherFactory(
            concurrency=self.concurrency,
            streaming=self.streaming,
            direct_copy=self.direct_copy,
            dry_run=self.dry_run,
            force=self.force,
        )
        self.aggregate_size = sum([self.science_ancillary.aggregate_size(report) for report in reports])

        # First we must prepare the fetchers for the location report
        # this itertools.chain is a recipe from the itertools documentation for flattening a mapping
        all_fetchers = list(
            itertools.chain.from_iterable(report.fetchers(self.science_ancillary, factory) for report in reports)
        )

        # now we have enough stuff to proceed with generating the plan
        return self.calculate_plan(factory, all_fetchers)

    @staticmethod
    def calculate_plan(factory: FetcherFactory, all_fetchers: List[FileFetcher]):
        """
        Calculates the fetch plan, given a list of fetchers to aggregate
        and a factory for generating fetchers.

        :param factory:
        :param all_fetchers:
        :return:
        """
        # The next step is to group like with like in terms of file fetcher resources.
        # We do this by creating a group each time we find a fetcher that
        # does not share resources with a "prototypical instance" from each
        # of the other groups.
        fetcher_classes = {}
        for fetcher in all_fetchers:
            found = False
            for class_prototype in fetcher_classes.keys():
                if fetcher.shares_resources_with(class_prototype):
                    found = True
                    fetcher_classes[class_prototype].append(fetcher)

            if not found:
                # if we made it here, we did not find a class prototype we share resources with
                # therefore, we are the prototype of a new class
                fetcher_classes[fetcher] = [fetcher]

        # now that we have sorted the fetchers into categories based on which one share resources
        # we can assume that within each category we want to see a certain kind of parallelism
        plans = []
        for fetch_class in fetcher_classes.values():
            plans.append(factory.fetch_plan(fetch_class))

        # now the overall plan is basically to execute each of those plans in parallel
        return factory.fetch_plan(plans, len(plans))

    @staticmethod
    def parse_commandline(args: List[str]) -> "FetchContext":
        """
        Parse the command line to build a FetchContext that handles options,
        if any

        :param args:
        :return:
        """
        parser = FetchContext.arg_parser()
        namespace = parser.parse_args(args)

        # determine the locators
        location_files: List[Locator] = [FileLocator(Path(file)) for file in namespace.location_files]
        service_locators: List[Locator] = [ServiceLocator(spl) for spl in namespace.product_locators]

        locators = location_files + service_locators

        # we can no longer rely on the argument parser to enforce this constraint, so we have to do it
        # ourselves here
        if len(locators) == 0:
            parser.error("You must supply at least one science product locator or location file")

        return FetchContext(
            locators,
            namespace.streaming,
            namespace.direct_copy,
            namespace.dry_run,
            ForceMode(namespace.force),
            namespace.science_ancillary,
            namespace.concurrency,
            namespace.progress_reporter,
        )

    @staticmethod
    def arg_parser() -> argparse.ArgumentParser:
        """
        Build and return an argument parser with the command line options
        for data-fetcher. This is static because Sphinx needs it to build
        the docs.

        Note difference from previous implementations: Capo profile,
        when needed, will be obtained from the environment;
        files will be retrieved to the current directory.

        Normally, whether to use the NGAS direct-copy plugin to retrieve
        a file is decided on the basis of a file's execution site,
        location, and cluster; the --streaming and --direct-copy flag can
        force one method or the other. If direct-copy is specified but
        not possible, an error will be thrown.

        :return: an argparse 'parser' with command line options for productfetcher.
        """

        default_threads = CapoConfig().settings("edu.nrao.workspaces.ProductFetcherSettings").defaultThreadsPerHost

        parser = argparse.ArgumentParser(
            description=_PROLOGUE,
            formatter_class=argparse.RawTextHelpFormatter,
        )

        parser.add_argument(
            CLIParam.CONCURRENCY.value,
            action="store",
            dest="concurrency",
            help="maximum threads per plan (set to 1 for no threading)",
            type=int,
            default=default_threads,
        )

        parser.add_argument(
            CLIParam.SPL.value, action="append", dest="product_locators", help="product locator to download", default=[]
        )
        parser.add_argument(
            CLIParam.FILE.value,
            action="append",
            dest="location_files",
            help="product locator report (in JSON)",
            default=[],
        )

        dry_or_force_options = parser.add_mutually_exclusive_group(required=False)
        dry_or_force_options.add_argument(
            CLIParam.DRY.value,
            action="store_true",
            dest="dry_run",
            help="dry run; do not fetch product",
        )

        dry_or_force_options.add_argument(
            CLIParam.FORCE.value,
            action="store_true",
            dest="force",
            help="overwrite existing file(s) at dest",
        )

        copy_or_stream_options = parser.add_mutually_exclusive_group(required=False)
        copy_or_stream_options.add_argument(
            CLIParam.DIRECT_COPY.value,
            action="store_true",
            dest="direct_copy",
            help="force direct-copy from NGAS",
        )
        copy_or_stream_options.add_argument(
            CLIParam.STREAMING.value,
            action="store_true",
            dest="streaming",
            help="force streaming from NGAS",
        )

        ancillary_options = parser.add_mutually_exclusive_group(required=False)
        ancillary_options.add_argument(
            CLIParam.ALSO_ANCILLARIES.value,
            action="store_const",
            dest="science_ancillary",
            const=FetchBothScienceAndAncillary(),
            help="fetch both the supplied product and its ancillary products, if any",
        )
        ancillary_options.add_argument(
            CLIParam.ONLY_ANCILLARIES.value,
            action="store_const",
            dest="science_ancillary",
            const=FetchAncillaryOnly(),
            help="only fetch the ancillary products of the supplied product",
        )
        ancillary_options.add_argument(
            CLIParam.ONLY_SCIENCE.value,
            action="store_const",
            dest="science_ancillary",
            const=FetchScienceOnly(),
            help="fetch only the supplied product, not any ancillary products (default behavior)",
        )

        progress_options = parser.add_mutually_exclusive_group(required=False)
        progress_options.add_argument(
            CLIParam.VERBOSE.value,
            action="store_const",
            dest="progress_reporter",
            const=VerboseProgressReporter,
            help="verbosely describe files being fetched",
        )
        progress_options.add_argument(
            CLIParam.PROGRESS.value,
            action="store_const",
            dest="progress_reporter",
            const=ProgressBarReporter,
            help="show a progress bar for files being fetched (default)",
        )
        progress_options.add_argument(
            CLIParam.QUIET.value,
            action="store_const",
            dest="progress_reporter",
            const=ProgressNonReporter,
            help="show no output (quiet mode)",
        )
        parser.set_defaults(science_ancillary=FetchScienceOnly(), progress_reporter=ProgressBarReporter)

        return parser

    def make_progress_reporter(self):
        return self.progress_reporter(self.aggregate_size)


def main(args=None):
    """CLI entry point"""
    try:
        # parse the arguments
        context = FetchContext.parse_commandline(args)

        # generate the plan
        plan = context.generate_plan()

        # perform the fetch
        plan.fetch(context.make_progress_reporter())

    except FetchError as ex:
        ex.terminate_fetch()
        sys.exit(1)


if __name__ == "__main__":
    main()
