#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for our various file fetchers """

# pylint: disable=C0103, E0401, E0402, W0621

from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from productfetcher.exceptions import (
    NormalModeFileExistsError,
    RetryableFetchError,
    UnsatisfiableConstraint,
)
from productfetcher.fetchers import (
    ForceMode,
    NgasDirectCopyFileFetcher,
    NgasStreamingFileFetcher,
    OracleXmlFetcher,
    RetryableFileFetcher,
)
from productfetcher.interfaces import FileFetcher, LocatedFile
from productfetcher.locations import Cluster, Location, NgasFile, NgasServer, OracleXml


class FailNTimesFileFetcher(FileFetcher):
    """A failin' fetcher. This dog don't fetch."""

    def __init__(self, n=3):
        super().__init__()
        self.failures = n

    @property
    def file(self) -> LocatedFile:
        """
        (Never called)

        :return:
        """

    def do_fetch(self) -> Path:
        """
        Try to fetch N times.

        :return:
        """
        if self.failures > 0:
            self.failures -= 1
            raise RetryableFetchError("testing failure " + str(self.failures))
        return Path.home()

    def shares_resources_with(self, other: "FileFetcher") -> bool:
        """
        (Never called)

        :return:
        """


@pytest.fixture
def ngas_file(tmp_path) -> NgasFile:
    """
    A fake NGAS file for our tests

    :param tmp_path:
    :return:
    """
    return NgasFile(
        "fake_ngas_id",
        tmp_path.name,
        "ASDMBinary",
        -1,
        "cksum",
        1,
        100,
        NgasServer("localhost", Location.DSOC, Cluster.DSOC),
    )


@pytest.fixture
def alma_file(tmp_path) -> OracleXml:
    """
    A fake ALMA SDM table for our tests

    :param tmp_path:
    :return:
    """
    return OracleXml("fake_uid", "fake_table", tmp_path.name, "relative_path", 100)


def test_force_mode(resource_path_root):
    """
    Does force mode overwrite an existing product?

    :param resource_path_root:
    :return:
    """
    # check files that do exist:
    assert ForceMode.FORCE.check_destination(resource_path_root / "location_files" / "alma-execblock.json") is None
    with pytest.raises(NormalModeFileExistsError):
        ForceMode.NORMAL.check_destination(resource_path_root / "location_files" / "alma-execblock.json")

    # check files that do not exist
    assert ForceMode.FORCE.check_destination(Path("/this/is/a/bogus/path")) is None
    assert ForceMode.NORMAL.check_destination(Path("/this/is/a/bogus/path")) is None


def test_ngas_direct_copy(ngas_file: NgasFile, tmp_path, capsys):
    """
    If we can direct copy, we should try to do so; if we can't, attempt to do so
     should throw exception

    :param ngas_file: what we're trying to fetch
    :param tmp_path: our temporary dir
    :param capsys:
    :return:
    """
    # make sure that if we could direct copy, we try to
    with patch("productfetcher.locations.NgasServer.can_direct_copy_to", return_value=True):
        with patch("productfetcher.ngas.NgasConnection.direct_copy", return_value=None):
            expected_path = Path(tmp_path) / "ASDMBinary"
            generated_path = tmp_path.parent / NgasDirectCopyFileFetcher(ngas_file).do_fetch()
            assert generated_path == expected_path

    # contrarily, if we cannot, we should get an exception
    with patch("productfetcher.locations.NgasServer.can_direct_copy_to", return_value=False):
        with patch("productfetcher.ngas.NgasConnection.direct_copy", return_value=None):
            with pytest.raises(UnsatisfiableConstraint):
                NgasDirectCopyFileFetcher(ngas_file).do_fetch()

    capsys.readouterr()


def test_ngas_streaming(ngas_file: NgasFile, tmp_path, capsys):
    # make sure that if we could direct copy, we try to
    with patch("productfetcher.ngas.NgasConnection.stream", return_value=None):
        expected_path = tmp_path / "ASDMBinary"
        generated_path = tmp_path.parent / NgasStreamingFileFetcher(ngas_file).do_fetch()
        assert generated_path == expected_path

    capsys.readouterr()


def test_alma_xml(alma_file: OracleXml, tmp_path, capsys):
    mock_engine = MagicMock()
    mock_engine.execute.return_value = [["this is the file content"]]
    with patch("productfetcher.fetchers.OracleXmlFetcher.get_engine", return_value=mock_engine):
        expected_path = tmp_path / "relative_path"
        generated_path = tmp_path.parent / OracleXmlFetcher(alma_file).do_fetch()
        assert generated_path == expected_path

    capsys.readouterr()


# ----------------------------------------------------------------------------
#
#      R e t r y i n g   T e s t s
#
# ----------------------------------------------------------------------------


def test_retrying_succeeds(capsys):
    # two failures is OK
    with patch("time.sleep"):
        fail_twice = MagicMock(wraps=FailNTimesFileFetcher(2))
        result = RetryableFileFetcher(fail_twice).do_fetch()
        assert result == Path.home()
        assert fail_twice.do_fetch.call_count == 3
    capsys.readouterr()


def test_retrying_fails(capsys):
    # three failures is too many
    fail_twice = MagicMock(wraps=FailNTimesFileFetcher(3))
    with patch("time.sleep"):
        with pytest.raises(RetryableFetchError):
            RetryableFileFetcher(fail_twice).do_fetch()
    capsys.readouterr()
